import os
from contextlib import contextmanager
try:
    from unittest.mock import patch
except ImportError:
    try:
        from mock import patch
    except ImportError as e:
        raise e

import logging
from glob import glob
from numpy import array


def create_directory(directory):
    if not os.path.isdir(directory):
        os.makedirs(directory)

@contextmanager
def stdout_disabled():
    devnull = open(os.devnull, 'w')
    with patch('sys.stdout', devnull):
        yield

@contextmanager
def all_logging_disabled():
    """
    A context manager that will prevent any logging messages triggered during
    the body from being processed.
    """
    # two kind-of hacks here:
    #    * can't get the highest logging level in effect => delegate to the user
    #    * can't get the current module-level override => use an undocumented
    #       (but non-private!) interface

    previous_level = logging.root.manager.disable

    logging.disable(logging.CRITICAL)

    try:
        yield
    finally:
        logging.disable(previous_level)

def find_data_file(file):
    here = os.path.dirname(os.path.abspath(__file__))
    data_file = os.path.join(here, '..', 'data', file)

    if '*' in data_file:
        files = sorted(glob(data_file))
        if len(files) > 0:
            return files

    if not os.path.exists(data_file):
        data_file = os.path.join(here, 'data', file)
        if '*' in data_file:
            data_file = sorted(glob(data_file))

    return data_file

def ESPRESSO_ADC_issues():
    adc_file = find_data_file('obs_affected_ADC_issues.dat')
    lines = [line.strip() for line in open(adc_file).readlines()]
    file_roots = [line.split()[1] for line in lines if not line.startswith('#')]
    return array(file_roots)

def ESPRESSO_cryostat_issues():
    cryostat_file = find_data_file('obs_affected_blue_cryostat_issues.dat')
    lines = [line.strip() for line in open(cryostat_file).readlines()]
    file_roots = [line.split()[1] for line in lines if not line.startswith('#')]
    return array(file_roots)



import os
from dataclasses import dataclass, field
from typing import Union
from functools import partial
from glob import glob
from datetime import datetime, timezone
import numpy as np

from astropy import units

from .setup_logger import logger
from .config import return_self
from .translations import translate
from .dace_wrapper import get_observations, get_arrays
from .dace_wrapper import do_download_ccf, do_download_s1d, do_download_s2d
from .simbad_wrapper import simbad
from .stats import wmean, wrms
from .binning import binRV


@dataclass
class RV:
    """
    A class holding RV observations

    Examples:
        >>> s = RV('Proxima')

    Attributes:
        star (str):
            The name of the star
        N (int):
            Total number of observations
        instruments (list):
            List of instruments for which there are RVs. Each instrument is also
            stored as an attribute (e.g. `self.CORALIE98` or `self.HARPS`)
        simbad (simbad):
            Information on the target from Simbad
    """
    star: str
    instrument: str = field(init=True, repr=False, default=None)
    N: int = field(init=False, repr=True)
    verbose: bool = field(init=True, repr=False, default=True)
    do_maxerror: Union[bool, float] = field(init=True, repr=False, default=False)
    do_secular_acceleration: bool = field(init=True, repr=False, default=True)
    do_sigma_clip: bool = field(init=True, repr=False, default=False)
    do_adjust_means: bool = field(init=True, repr=False, default=True)
    #
    _child: bool = field(init=True, repr=False, default=False)
    _did_secular_acceleration: bool = field(init=False, repr=False, default=False)
    _did_sigma_clip: bool = field(init=False, repr=False, default=False)
    _did_adjust_means: bool = field(init=False, repr=False, default=False)
    _raise_on_error: bool = field(init=True, repr=False, default=True)

    def __repr__(self):
        if self.time.size == self.mtime.size:
            return f"RV(star='{self.star}', N={self.N})"
        else:
            nmasked = self.N - self.mtime.size
            return f"RV(star='{self.star}', N={self.N}, masked={nmasked})"

    def __post_init__(self):
        self.__star__ = translate(self.star)

        if not self._child:
            try:
                self.simbad = simbad(self.__star__)
            except ValueError as e:
                logger.error(e)

            if self.verbose:
                logger.info(f'querying DACE for {self.__star__}...')
            try:
                self.dace_result = get_observations(self.__star__, self.instrument,
                                                    verbose=self.verbose)
            except ValueError as e:
                if self._raise_on_error:
                    raise e
                else:
                    self.time = np.array([])
                    self.instruments = []
                    self.units = ''
                    return


            # store the date of the last DACE query
            time_stamp = datetime.now(timezone.utc)  #.isoformat().split('.')[0]
            self._last_dace_query = time_stamp

        self.units = 'm/s'

        # build children
        if not self._child:
            arrays = get_arrays(self.dace_result, verbose=self.verbose)
            for (inst, pipe, mode), data in arrays:
                child = RV.from_dace_data(self.star, inst, pipe, mode, data, _child=True)
                setattr(self, inst, child)

        # build joint arrays
        if not self._child:
            #! sorted?
            self.instruments = sorted(list(self.dace_result.keys()))
            # self.pipelines =
            # "observatory" (or instrument id)
            self.obs = np.concatenate(
                [np.full(getattr(self, inst).N, i+1) for i, inst in enumerate(self.instruments)],
                dtype=int
            )
            # mask
            self.mask = np.full_like(self.obs, True, dtype=bool)
            # all other quantities
            self._build_arrays()

        # do clip_maxerror, secular_acceleration, sigmaclip, adjust_means
        if not self._child:
            if self.do_maxerror:
                self.clip_maxerror(self.do_maxerror)

            if self.do_secular_acceleration:
                self.secular_acceleration()

            if self.do_sigma_clip:
                self.sigmaclip()

            if self.do_adjust_means:
                self.adjust_means()


    def reload(self):
        self._did_secular_acceleration = False
        self._did_sigma_clip = False
        self._did_adjust_means = False
        self.__post_init__()

    def snapshot(self):
        import pickle
        from datetime import datetime
        ts = datetime.now().timestamp()
        star_name = self.star.replace(' ', '')
        file = f'{star_name}_{ts}.pkl'
        pickle.dump(self, open(file, 'wb'), protocol=0)
        if self.verbose:
            logger.info(f'Saved snapshot to {file}')

    @property
    def N(self):
        """Total number of observations"""
        return self.time.size

    @N.setter
    def N(self, value):
        if not isinstance(value, property):
            logger.error('Cannot set N directly')

    @property
    def NN(self):
        """ Total number of observations per instrument """
        return {inst: getattr(self, inst).N for inst in self.instruments}

    @property
    def N_nights(self):
        """ Number of individual nights """
        return binRV(self.mtime, None, None, binning_bins=True).size - 1

    @property
    def NN_nights(self):
        return {inst: getattr(self, inst).N_nights for inst in self.instruments}

    @property
    def mtime(self):
        return self.time[self.mask]

    @property
    def mvrad(self):
        return self.vrad[self.mask]

    @property
    def msvrad(self):
        return self.svrad[self.mask]

    @property
    def instrument_array(self):
        return np.concatenate([[i] * n for i, n in self.NN.items()])

    @property
    def rms(self):
        """ Weighted rms of the (masked) radial velocities """
        if self.mask.sum() == 0:  # only one point
            return np.nan
        else:
            return wrms(self.vrad[self.mask], self.svrad[self.mask])

    @property
    def sigma(self):
        """ Average error bar """
        if self.mask.sum() == 0:  # only one point
            return np.nan
        else:
            return self.svrad[self.mask].mean()

    error = sigma  # alias!

    @classmethod
    def from_dace_data(cls, star, inst, pipe, mode, data, **kwargs):
        s = cls(star, **kwargs)
        #
        ind = np.argsort(data['rjd'])
        # time, RVs, uncertainties
        s.time = data['rjd'][ind]
        s.vrad = data['rv'][ind]
        s.svrad = data['rv_err'][ind]
        # mask
        s.mask = np.full_like(s.time, True, dtype=bool)
        s.mask[np.isnan(s.svrad)] = False
        # all other quantities
        s._quantities = []
        for arr in data.keys():
            if arr not in ('rjd', 'rv', 'rv_err'):
                if arr == 'mask':
                    # change name mask -> ccf_mask
                    setattr(s, 'ccf_mask', data[arr][ind])
                    s._quantities.append('ccf_mask')
                else:
                    setattr(s, arr, data[arr][ind])
                    s._quantities.append(arr)
        #
        s.instruments = [inst]
        s.pipelines = [pipe]
        s.modes = [mode]

        return s

    @classmethod
    def from_arrays(cls, star, time, vrad, svrad, inst, *args, **kwargs):
        s = cls(star, _child=True)
        time, vrad, svrad = map(np.atleast_1d, (time, vrad, svrad))

        if time.size != vrad.size:
            logger.error(f'wrong dimensions: time({time.size}) != vrad({vrad.size})')
            raise ValueError from None
        if time.size != svrad.size:
            logger.error(f'wrong dimensions: time({time.size}) != svrad({svrad.size})')
            raise ValueError from None

        # time, RVs, uncertainties
        s.time = time
        s.vrad = vrad
        s.svrad = svrad
        # mask
        s.mask = kwargs.get('mask', np.full_like(s.time, True, dtype=bool))

        s.instruments = [inst]

        return s

    @classmethod
    def from_snapshot(cls, file):
        import pickle
        from datetime import datetime
        assert file.endswith('.pkl'), 'expected a .pkl file'
        star, timestamp = file.replace('.pkl', '').split('_')
        dt = datetime.fromtimestamp(float(timestamp))
        logger.info(f'Reading snapshot of {star} from {dt}')
        return pickle.load(open(file, 'rb'))

    def _check_instrument(self, instrument):
        if instrument is None:
            return self.instruments
        if instrument in self.instruments:
            return [instrument]
        if any([instrument in inst for inst in self.instruments]):
            return [inst for inst in self.instruments if instrument in inst]
        

    def _build_arrays(self):
        """ build all concatenated arrays of `self` from each of the `.inst`s """
        if self._child:
            return
        # time
        self.time = np.concatenate(
            [getattr(self, inst).time for inst in self.instruments]
        )
        # RVs
        self.vrad = np.concatenate(
            [getattr(self, inst).vrad for inst in self.instruments]
        )
        # uncertainties
        self.svrad = np.concatenate(
            [getattr(self, inst).svrad for inst in self.instruments]
        )

        # mask
        self.mask = np.concatenate(
            [getattr(self, inst).mask for inst in self.instruments]
        )

        # all other quantities
        self._quantities = getattr(self, self.instruments[0])._quantities
        if len(self.instruments) > 1:
            for inst in self.instruments[1:]:
                self._quantities = np.intersect1d(self._quantities, getattr(self, inst)._quantities)

        for q in self._quantities:
            if q not in ('rjd', 'rv', 'rv_err'):
                arr = np.concatenate(
                    [getattr(getattr(self, inst), q) for inst in self.instruments]
                )
                setattr(self, q, arr)


    def download_ccf(self, instrument=None, limit=None, directory=None):
        """ Download CCFs from DACE

        Args:
            instrument (str): Specific instrument for which to download data
            limit (int): Maximum number of files to download.
            directory (str): Directory where to store data.
        """
        if directory is None:
            directory = f'{self.star}_downloads'

        if instrument is None:
            files = [file for file in self.raw_file if file.endswith('.fits')]
        else:
            instrument = self._check_instrument(instrument)
            files = []
            for inst in instrument:
                files += list(getattr(self, inst).raw_file)

        do_download_ccf(files[:limit], directory)

    def download_s1d(self, instrument=None, limit=None, directory=None):
        """ Download S1Ds from DACE

        Args:
            instrument (str): Specific instrument for which to download data
            limit (int): Maximum number of files to download.
            directory (str): Directory where to store data.
        """
        if directory is None:
            directory = f'{self.star}_downloads'

        if instrument is None:
            files = [file for file in self.raw_file if file.endswith('.fits')]
        else:
            instrument = self._check_instrument(instrument)
            files = []
            for inst in instrument:
                files += list(getattr(self, inst).raw_file)

        do_download_s1d(files[:limit], directory)

    def download_s2d(self, instrument=None, limit=None, directory=None):
        """ Download S2Ds from DACE

        Args:
            instrument (str): Specific instrument for which to download data
            limit (int): Maximum number of files to download.
            directory (str): Directory where to store data.
        """
        if directory is None:
            directory = f'{self.star}_downloads'

        if instrument is None:
            files = [file for file in self.raw_file if file.endswith('.fits')]
        else:
            instrument = self._check_instrument(instrument)
            files = []
            for inst in instrument:
                files += list(getattr(self, inst).raw_file)

        extracted_files = do_download_s2d(files[:limit], directory)


    from .plots import plot, plot_fwhm, plot_bis, plot_rhk
    from .plots import gls, gls_fwhm, gls_bis, gls_rhk
    from .reports import report

    from .instrument_specific import known_issues


    def remove_instrument(self, instrument):
        """ Remove all observations from one instrument
        
        Args:
            instrument (str): The instrument for which to remove observations.
        
        Note:
            A common name can be used to remove observations for several subsets
            of a given instrument. For example

            ```py
            s.remove_instrument('HARPS')
            ```

            will remove observations from `HARPS03` and `HARPS15`, if they
            exist. But

            ```py
            s.remove_instrument('HARPS03')
            ```

            will remove observations from the specific subset.
        """
        if instrument not in self.instruments:
            logger.error(f"No data from instrument '{instrument}'")
            logger.info(f'available: {self.instruments}')
            return

        ind = self.instruments.index(instrument) + 1
        remove = np.where(self.obs == ind)
        self.obs = np.delete(self.obs, remove)
        self.obs[self.obs > ind] -= 1
        #
        self.time = np.delete(self.time, remove)
        self.vrad = np.delete(self.vrad, remove)
        self.svrad = np.delete(self.svrad, remove)
        #
        self.mask = np.delete(self.mask, remove)
        #
        # all other quantities
        for q in self._quantities:
            if q not in ('rjd', 'rv', 'rv_err'):
                new = np.delete(getattr(self, q), remove)
                setattr(self, q, new)
        #
        self.instruments.remove(instrument)
        #
        delattr(self, instrument)

        if self.verbose:
            logger.info(f"Removed observations from '{instrument}'")

        if return_self:
            return self

    def remove_point(self, index):
        """ Remove individual observations at a given index (or indices)
        
        Args:
            index (int, list, ndarray):
                Single index, list, or array of indices to remove.
        """
        index = np.atleast_1d(index)
        try:
            instrument_index = self.obs[index]
            instrument = np.array(self.instruments)[instrument_index - 1]
        except IndexError:
            logger.errors(f'index {index} is out of bounds for N={self.N}')
            return

        self.mask[index] = False
        self._propagate_mask_changes()
        # for i, inst in zip(index, instrument):
        #     index_in_instrument = i - (self.obs < instrument_index).sum()
        #     getattr(self, inst).mask[index_in_instrument] = False
        if return_self:
            return self

    def _propagate_mask_changes(self):
        """ link self.mask with each self.`instrument`.mask """
        masked = np.where(~self.mask)[0]
        for m in masked:
            inst = self.instruments[self.obs[m] - 1]
            n_before = (self.obs < self.obs[m]).sum()
            getattr(self, inst).mask[m - n_before] = False

    def secular_acceleration(self, epoch=55500, plot=False):
        """
        Remove secular acceleration from RVs

        Args:
            epoch (float):
                The reference epoch (DACE uses 55500, 31/10/2010)
            instruments (bool or collection of str):
                Only remove secular acceleration for some instruments, or for all 
                if `instruments=True`
            plot (bool):
                Show a plot of the RVs with the secular acceleration
        """
        if self._did_secular_acceleration:  # don't do it twice
            return
        
        try:
            self.simbad
        except AttributeError:
            if self.verbose:
                logger.error('no information from simbad, cannot remove secular acceleration')
            return

        as_yr = units.arcsec / units.year
        mas_yr = units.milliarcsecond / units.year
        mas = units.milliarcsecond

        π = self.simbad.plx_value * mas
        d = π.to(units.pc, equivalencies=units.parallax())
        μα = self.simbad.pmra * mas_yr
        μδ = self.simbad.pmdec * mas_yr
        μ = μα**2 + μδ**2
        sa = (μ * d).to(units.m / units.second / units.year,
                        equivalencies=units.dimensionless_angles())
        sa = sa.value

        if self.verbose:
            logger.info('removing secular acceleration from RVs')

        if self.units == 'km/s':
            sa /= 1000

        if self._child:
            self.vrad = self.vrad - sa * (self.time - epoch) / 365.25
        else:
            for inst in self.instruments:
                if 'HIRES' in inst:  # never remove it from HIRES...
                    continue
                if 'NIRPS' in inst:  # never remove it from NIRPS...
                    continue

                s = getattr(self, inst)
                s.vrad = s.vrad - sa * (s.time - epoch) / 365.25
            
            self._build_arrays()

        self._did_secular_acceleration = True
        if return_self:
            return self

    def sigmaclip(self, sigma=3):
        """ Sigma-clip RVs """
        if self._child or self._did_sigma_clip:
            return
        from scipy.stats import sigmaclip as dosigmaclip
        result = dosigmaclip(self.vrad, low=sigma, high=sigma)
        n = self.vrad.size - result.clipped.size

        if self.verbose and n > 0:
            s = 's' if (n == 0 or n > 1) else ''
            logger.warning(f'sigma-clip RVs will remove {n} point' + s)

        ind = (self.vrad > result.lower) & (self.vrad < result.upper)

        # check if going to remove all observations from one instrument
        if n in self.NN.values(): # all observations
            insts = np.unique(self.instrument_array[~ind])
            if insts.size == 1: # of the same instrument?
                if self.verbose:
                    logger.warning(f'would remove all observations from {insts[0]}, skipping')
                if return_self:
                    return self

        self.mask[~ind] = False
        self._propagate_mask_changes()
        if return_self:
            return self

    def clip_maxerror(self, maxerror:float, plot=False):
        """ Mask out points with RV error larger than a given value
        
        Args:
            maxerror (float): Maximum error to keep.
            plot (bool): Whether to plot the masked points.
        """
        if self._child:
            return

        self.maxerror = maxerror
        above = self.svrad > maxerror
        n = above.sum()
        self.mask[above] = False

        if self.verbose and above.sum() > 0:
            s = 's' if (n == 0 or n > 1) else ''
            logger.warning(f'clip_maxerror ({maxerror} {self.units}) removed {n} point' + s)

        self._propagate_mask_changes()
        if return_self:
            return self

    def bin(self):
        """ Nightly bin the observations """
        for inst in self.instruments:
            s = getattr(self, inst)
            tb, vb, svb = binRV(s.time, s.vrad, s.svrad)
            s.vrad = vb
            s.svrad = svb

            for q in s._quantities:
                if getattr(s, q).dtype != np.float64:
                    if self.verbose:
                        logger.warning(f"{inst}, skipping {q} in binning")
                    continue
                if np.isnan(getattr(s, q)).all():
                    yb = np.full_like(tb, np.nan)
                    setattr(s, q, yb)
                elif q + '_err' in s._quantities:
                    _, yb, eb = binRV(s.time, getattr(s, q), getattr(s, q + '_err'))
                    setattr(s, q, yb)
                    setattr(s, q + '_err', eb)
                elif not q.endswith('_err'):
                    try:
                        _, yb = binRV(s.time, getattr(s, q), stat=np.nanmean, tstat=np.nanmean)
                        setattr(s, q, yb)
                    except TypeError:
                        pass

            s.time = tb
            s.mask = np.full(tb.shape, True)
        
        self._build_arrays()

    def adjust_means(self, just_rv=False):
        if self._child or self._did_adjust_means:
            return

        others = ('fwhm', 'bispan', )
        for inst in self.instruments:
            s = getattr(self, inst)

            if s.N == 1:
                if self.verbose:
                    msg = (f'only 1 observation for {inst}, '
                           'adjust_means will set it exactly to zero')
                    logger.warning(msg)
                s.rv_mean = s.vrad[0]
                s.vrad = np.zeros_like(s.time)
                continue

            s.rv_mean = wmean(s.mvrad, s.msvrad)
            s.vrad -= s.rv_mean
            if self.verbose:
                logger.info(f'subtracted weighted average from {inst:10s}: ({s.rv_mean:.3f} {self.units})')
            if just_rv:
                continue
            # log_msg = 'same for '
            for i, other in enumerate(others):
                y, ye = getattr(s, other), getattr(s, other + '_err')
                m = wmean(y[s.mask], ye[s.mask])
                setattr(s, f'{other}_mean', m)
                setattr(s, other, getattr(s, other) - m)
                # log_msg += other
                # if i < len(others) - 1:
                #     log_msg += ', '
            
            # if self.verbose:
            #     logger.info(log_msg)

        self._build_arrays()
        self._did_adjust_means = True
        if return_self:
            return self

    #

    def save(self, directory=None, instrument=None, full=False):
        """ Save the observations in .rdb files.

        Args:
            directory (str, optional):
                Directory where to save the .rdb files.
            instrument (str, optional):
                Instrument for which to save observations.
            full (bool, optional): 
                Whether to save just RVs and errors (False) or more indicators (True).
        """
        star_name = self.star.replace(' ', '')

        files = []

        for inst in self.instruments:
            if instrument is not None:
                if instrument not in inst:
                    continue

            file = f'{star_name}_{inst}.rdb'
            files.append(file)

            if directory is not None:
                os.makedirs(directory, exist_ok=True)
                file = os.path.join(directory, file)

            _s = getattr(self, inst)

            if full:
                d = np.c_[
                    _s.mtime, _s.mvrad, _s.msvrad,
                    _s.fwhm[_s.mask], _s.fwhm_err[_s.mask]
                ]
                header =  'bjd\tvrad\tsvrad\tfwhm\tsfwhm\n'
                header += '---\t----\t-----\t----\t-----'
            else:
                d = np.c_[_s.mtime, _s.mvrad, _s.msvrad]
                header = 'bjd\tvrad\tsvrad\n---\t----\t-----'
            
            np.savetxt(file, d, fmt='%9.5f', header=header, delimiter='\t', comments='')

            if self.verbose:
                logger.info(f'saving to {file}')
        
        return files

    #
    def run_lbl(self, instrument=None, data_dir=None, 
                skysub=False, tell=False, limit=None, **kwargs):
        from .lbl_wrapper import run_lbl, NIRPS_create_telluric_corrected_S2D

        if instrument is None:
            instruments = self.instruments
        else:
            if instrument not in self.instruments:
                if any([instrument in i for i in self.instruments]):
                    instrument = [i for i in self.instruments if instrument in i]
                else:
                    logger.error(f"No data from instrument '{instrument}'")
                    logger.info(f'available: {self.instruments}')
                    return
            
            if isinstance(instrument, str):
                instruments = [instrument]
            else:
                instruments = instrument

        for instrument in instruments:
            if self.verbose:
                logger.info(f'gathering files for {instrument}')
            files = getattr(self, instrument).raw_file
            files = map(os.path.basename, files)
            if skysub:
                files = [file.replace('.fits', '_S2D_SKYSUB_A.fits') for file in files]
            else:
                files = [file.replace('.fits', '_S2D_A.fits') for file in files]

            if data_dir is None:
                data_dir = f'{self.star}_downloads'

            files = [os.path.join(data_dir, file) for file in files]
            exist = [os.path.exists(file) for file in files]
            if not all(exist):
                logger.error(f"not all required files exist in {data_dir}")
                logger.error(f"missing {np.logical_not(exist).sum()} / {len(files)}")

                from distutils.util import strtobool
                go_on = input('continue? (y/N) ')
                if go_on == '' or not bool(strtobool(go_on)):
                    return

                files = list(np.array(files)[exist])

            # deal with NIRPS telluric correction
            if 'NIRPS' in instrument and tell:
                if self.verbose:
                    logger.info('creating telluric-corrected S2D files')
                files = NIRPS_create_telluric_corrected_S2D(files[:limit])

            run_lbl(self, instrument, files[:limit], **kwargs)

    def load_lbl(self, instrument=None, tell=False):
        if hasattr(self, '_did_load_lbl') and self._did_load_lbl: # don't do it twice
            return
        from .lbl_wrapper import load_lbl

        if instrument is None:
            instruments = self.instruments
        else:
            if instrument not in self.instruments:
                if any([instrument in i for i in self.instruments]):
                    instrument = [i for i in self.instruments if instrument in i]
                else:
                    logger.error(f"No data from instrument '{instrument}'")
                    logger.info(f'available: {self.instruments}')
                    return
            
            if isinstance(instrument, str):
                instruments = [instrument]
            else:
                instruments = instrument

        for inst in instruments:
            load_lbl(self, inst, tell=tell)

        self._did_load_lbl = True


def fit_sine(t, y, yerr, period='gls', fix_period=False):
    from scipy.optimize import leastsq
    if period == 'gls':
        from astropy.timeseries import LombScargle
        gls = LombScargle(t, y, yerr)
        freq, power = gls.autopower()
        period = 1 / freq[power.argmax()]

    if fix_period and period is None:
        logger.error('period is fixed but no value provided')
        return

    def sine(t, p):
        return p[0] * np.sin(2 * np.pi * t / p[1] + p[2]) + p[3]

    p0 = [y.ptp(), period, 0.0, 0.0]
    xbest, _ = leastsq(lambda p, t, y, ye: (sine(t, p) - y) / ye,
                       p0, args=(t, y, yerr))
    return xbest, partial(sine, p=xbest)


import os
import concurrent.futures
from collections import namedtuple
from tqdm import tqdm
# import numpy as np

from .setup_logger import logger
from .timeseries import RV

__all__ = ['ESPRESSO_GTO']

path = os.path.join(os.path.dirname(__file__), 'data')


def get_star(star, instrument=None):
    return RV(star, verbose=False, instrument=instrument, _raise_on_error=False)


class LazyRV:
    def __init__(self, stars: list, instrument: str = None):
        self.stars = stars
        if isinstance(self.stars, str):
            self.stars = [self.stars]
        self.instrument = instrument

    @property
    def N(self):
        return len(self.stars)

    def __repr__(self):
        return f"RV({self.N} stars)"

    def _get(self):
        result = []
        # use a with statement to ensure threads are cleaned up promptly
        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as pool:
            star_to_RV = {
                pool.submit(get_star, star, self.instrument): star 
                for star in self.stars
            }
            logger.info('Querying DACE...')
            pbar = tqdm(concurrent.futures.as_completed(star_to_RV),
                        total=self.N, unit='star')
            for future in pbar:
                star = star_to_RV[future]
                pbar.set_description(star)
                try:
                    result.append(future.result())
                except Exception:
                    print(f'{star} generated an exception')
        return result

    def __iter__(self):
        return self._get()

    def __call__(self):
        return self._get()


# sorted by spectral type
WG1_stars = [
    "HIP11533",  # F2
    "HD63077",  # F9
    "HD102365",  # G2
    "HD160691",  # G3
    "HD20794",  # G6
    "HD115617",  # G6
    "HD10700",  # G8
    "HD69830",  # G8
    "HD26965",  # K0
    "HD100623",  # K0
    "HD154088",  # K0
    "HD72673",  # K1
    "HD4628",  # K2
    "HD191408",  # K2
    "HD192310",  # K2
    "HD16160",  # K3
    "HD32147",  # K3
    "HD22496",  # K5
    "HIP93069",  # K5
    "HD209100",  # K5
    "HIP23708",  # K7
    "HD152606",  # K8
    "HD260655",  # M0
    "HIP40239",  # M0
    "HD304636",  # M0
    "HIP85647",  # M0
    "HD165222",  # M0
    "HD191849",  # M0
    "GJ191",  # M1
    "HD42581",  # M1
    "HIP42748",  # M1
    "HIP65859",  # M1
    "HIP86287",  # M1
    "HD176029",  # M1
    "GJ825",  # M1
    "HD225213",  # M2
    "GJ54",  # M2
    "HIP22627",  # M2
    "HIP51317",  # M2
    "HD119850",  # M2
    "GJ832",  # M2
    "HD217987",  # M2
    "GJ273",  # M3
    "GJ388",  # M3
    "HIP62452",  # M3
    "HIP67164",  # M3
    "HIP71253",  # M3
    "GJ628",  # M3
    "HIP85523",  # M3
    "HIP86214",  # M3
    "HIP92403",  # M3
    "HIP113020",  # M3
    "HIP1242",  # M4
    "GJ83.1",  # M4
    "HIP53020",  # M4
    "Ross128",  # M4
    "GJ699",  # M4
    "GJ1002",  # M5
    "GJ1061",  # M5
    "GJ3618",  # M5
    "Proxima",  # M5
    "GJ406",  # M6
]

ESPRESSO_GTO_nt = namedtuple('ESPRESSO_GTO', ['WG1', 'WG2', 'WG3'])
ESPRESSO_GTO = ESPRESSO_GTO_nt(
    WG1=LazyRV(WG1_stars, instrument='ESPRESSO'),
    WG2=LazyRV([], instrument='ESPRESSO'), # TODO
    WG3=LazyRV([], instrument='ESPRESSO'), # TODO
)
ESPRESSO_GTO.WG1.__doc__ = 'RV observations for all WG1 targets. Call ESPRESSO_GTO.WG1() to load them.'
ESPRESSO_GTO.WG2.__doc__ = 'RV observations for all WG2 targets. Call ESPRESSO_GTO.WG2() to load them.'
ESPRESSO_GTO.WG3.__doc__ = 'RV observations for all WG3 targets. Call ESPRESSO_GTO.WG3() to load them.'



from dataclasses import dataclass, field
import requests

from astropy.coordinates import SkyCoord
import pysweetcat


QUERY = """
SELECT basic.OID,
       RA,
       DEC,
       main_id,
       pmra,
       pmdec,
       plx_value,
       rvz_radvel,
       sp_type
FROM basic JOIN ident ON oidref = oid
WHERE id = '{star}';
"""

BV_QUERY = """
SELECT B, V from allfluxes
JOIN ident USING(oidref)
WHERE id = '{star}';
"""

IDS_QUERY = """
SELECT ids from ids
JOIN ident USING(oidref)
WHERE id = '{star}';
"""

def run_query(query):
    url = 'http://simbad.u-strasbg.fr/simbad/sim-tap/sync'
    response = requests.post(url,
                             data=dict(query=query,
                                       request='doQuery',
                                       lang='ADQL',
                                       format='text/plain',
                                       phase='run'))
    return response.content.decode()

def parse_table(table, cols=None, values=None):
    header = table.splitlines()[0].split('|')
    if cols is None:
        cols = list(map(str.strip, header))
    else:
        cols = cols + list(map(str.strip, header))
    if values is None:
        values = table.splitlines()[2].split('|')
    else:
        values = values + table.splitlines()[2].split('|')
    values = list(map(str.strip, values))
    values = [value.replace('"', '') for value in values]
    return cols, values


effective_temperatures = {
    'F0': 7350, 'F2': 7050, 'F3': 6850, 'F5': 6700, 'F6': 6550, 'F7': 6400, 'F8': 6300, 
    'G0': 6050, 'G1': 5930, 'G2': 5800, 'G5': 5660, 'G8': 5440,
    'K0': 5240, 'K1': 5110, 'K2': 4960, 'K3': 4800, 'K4': 4600, 'K5': 4400, 'K7': 4000,
    'M0': 3750, 'M1': 3700, 'M2': 3600, 'M3': 3500, 'M4': 3400, 'M5': 3200, 'M6': 3100, 'M7': 2900, 'M8': 2700,
}


class simbad:
    """
    A very simple wrapper around a TAP query to simbad for a given target. This
    class simply runs a few TAP queries and stores the result as attributes.

    Attributes:
        ra (float): right ascension
        dec (float): declination
        coords (SkyCoord): coordinates as a SkyCoord object
        main_id (str): main identifier
        plx_value (float): parallax
        rvz_radvel (float): radial velocity
        sp_type (str): spectral type
        B (float): B magnitude 
        V (float): V magnitude
        ids (list): list of identifiers
    """
    def __init__(self, star:str):
        """
        Args:
            star (str): The name of the star to query simbad
        """
        self.star = star
        try:
            table1 = run_query(query=QUERY.format(star=star))
            cols, values = parse_table(table1)

            table2 = run_query(query=BV_QUERY.format(star=star))
            cols, values = parse_table(table2, cols, values)

            table3 = run_query(query=IDS_QUERY.format(star=star))
            line = table3.splitlines()[2]
            self.ids = line.replace('"', '').replace('    ', ' ').replace('   ', ' ').replace('  ', ' ').split('|')
        except IndexError:
            raise ValueError(f'simbad query for {star} failed')

        for col, val in zip(cols, values):
            try:
                setattr(self, col, float(val))
            except ValueError:
                setattr(self, col, val)

        self.coords = SkyCoord(self.ra, self.dec, unit='deg')

        try:
            swc_data = pysweetcat.get_data()
            data = swc_data.find(star)
            if data is None:
                for id in self.ids:
                    data = swc_data.find(id)
                    if data is not None:
                        break
            if data is None:
                raise IndexError
            else:
                self.teff = data['teff']

        except IndexError:
            if self.sp_type[:2] in effective_temperatures:
                self.teff = effective_temperatures[self.sp_type[:2]]


    def __repr__(self):
        V = self.V
        sp_type = self.sp_type
        return f'{self.star} ({V=}, {sp_type=})'



def argsort_by_spectral_type(sptypes):
    STs = [f'{letter}{n}' for letter in ('F', 'G', 'K', 'M') for n in range(10)]
    order = {st: i for i, st in enumerate(STs)}
    indices = {i:st for i, st in zip(range(len(sptypes)), sptypes)}
    return [i[0] for i in sorted(indices.items(), key=lambda item: order[item[1]])]



import os
from functools import partial, partialmethod

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection
import mplcursors

from astropy.timeseries import LombScargle

from .setup_logger import logger
from . import config


def plot(self, ax=None, show_masked=False, time_offset=0, remove_50000=False, 
         tooltips=True, label=None, N_in_label=False, versus_n=False, show_histogram=False,
         **kwargs):
    """ Plot the RVs

    Args:
        ax (Axes, optional):
            Axis to plot to. Defaults to None.
        show_masked (bool, optional):
            Show masked points. Defaults to False.
        time_offset (int, optional):
            Value to subtract from time. Defaults to 0.
        remove_50000 (bool, optional):
            Whether to subtract 50000 from time. Defaults to False.
        tooltips (bool, optional):
            Show information upon clicking a point. Defaults to True.
        N_in_label (bool, optional):
            Show number of observations in legend. Defaults to False.
        versus_n (bool, optional):
            Plot RVs vs observation index instead of time. Defaults to False.
        show_histogram (bool, optional)
            Whether to show a panel with the RV histograms (per intrument).
            Defaults to False.

    Returns:
        Figure: the figure
        Axes: the axis
    """
    if self.N == 0:
        if self.verbose:
            logger.error('no data to plot')
        return

    if ax is None:
        if show_histogram:
            fig, (ax, axh) = plt.subplots(1, 2, constrained_layout=True, gridspec_kw={'width_ratios': [3, 1]})
        else:
            fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        if show_histogram:
            ax, axh = ax
        fig = ax.figure

    kwargs.setdefault('fmt', 'o')
    kwargs.setdefault('capsize', 0)
    kwargs.setdefault('ms', 4)

    if remove_50000:
        time_offset = 50000

    cursors = {}
    for inst in self.instruments:
        s = self if self._child else getattr(self, inst)

        if label is None:
            _label = f'{inst:10s} ({s.N})' if N_in_label else inst
        else:
            _label = label

        if versus_n:
            container = ax.errorbar(np.arange(1, s.mtime.size + 1),
                                    s.mvrad, s.msvrad, label=_label, picker=True, **kwargs)
        else:
            container = ax.errorbar(s.mtime - time_offset,
                                    s.mvrad, s.msvrad, label=_label, picker=True, **kwargs)

        if show_histogram:
            kw = dict(histtype='step', bins='doane', orientation='horizontal')
            hlabel = f'{s.mvrad.std():.2f} {self.units}'
            axh.hist(s.mvrad, **kw, label=hlabel)

        if tooltips:
            cursors[inst] = crsr = mplcursors.cursor(container, multiple=False)

            @crsr.connect("add")
            def _(sel):
                inst = sel.artist.get_label()
                _s = getattr(self, inst)
                vrad, svrad = _s.vrad[sel.index], _s.svrad[sel.index]
                sel.annotation.get_bbox_patch().set(fc="white")
                text = f'{inst}\n'
                text += f'BJD: {sel.target[0]:9.5f}\n'
                text += f'RV: {vrad:.3f} ± {svrad:.3f}'
                if fig.canvas.manager.toolmanager.get_tool('infotool').toggled:
                    text += '\n\n'
                    text += f'date: {_s.date_night[sel.index]}\n'
                    text += f'mask: {_s.ccf_mask[sel.index]}'
                sel.annotation.set_text(text)

        if show_masked:
            if versus_n:
                pass
            else:
                ax.errorbar(s.time[~s.mask] - time_offset, s.vrad[~s.mask], s.svrad[~s.mask],
                            **kwargs, color=container[0].get_color())

    if show_masked:
        if versus_n:
            pass
            # ax.errorbar(self.time[~self.mask] - time_offset, self.vrad[~self.mask], self.svrad[~self.mask],
            #             label='masked', fmt='x', ms=10, color='k', zorder=-2)
        else:
            ax.errorbar(self.time[~self.mask] - time_offset, self.vrad[~self.mask], self.svrad[~self.mask],
                        label='masked', fmt='x', ms=10, color='k', zorder=-2)

    ax.legend()
    if show_histogram:
        axh.legend()

    ax.minorticks_on()

    ax.set_ylabel(f'RV [{self.units}]')
    if versus_n:
        ax.set_xlabel('observation')
    else:
        if remove_50000:
            ax.set_xlabel('BJD - 2450000 [days]')
        else:
            ax.set_xlabel('BJD - 2400000 [days]')

    from matplotlib.backend_tools import ToolBase, ToolToggleBase
    tm = fig.canvas.manager.toolmanager

    class InfoTool(ToolToggleBase):
        description = "Show extra information about each observation"
        image = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data', 'info.svg'))
        # def enable(self, *args, **kwargs):
        #     self.figure.add_axes([1, 0, 0.3, 1])
        #     self.figure.canvas.draw_idle()

    from PIL import UnidentifiedImageError
    try:
        tm.add_tool("infotool", InfoTool)
        fig.canvas.manager.toolbar.add_tool(tm.get_tool("infotool"), "toolgroup")
        raise UnidentifiedImageError
    except AttributeError:
        pass
    except UnidentifiedImageError:
        pass


    if config.return_self:
        return self
    
    if show_histogram:
        return fig, (ax, axh)
    
    return fig, ax


def plot_quantity(self, quantity, ax=None, show_masked=False, time_offset=0, 
                  remove_50000=False, tooltips=False, N_in_label=False, **kwargs):
    if self.N == 0:
        if self.verbose:
            logger.error('no data to plot')
        return

    if not hasattr(self, quantity):
        logger.error(f"cannot find '{quantity}' attribute")
        return

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    kwargs.setdefault('fmt', 'o')
    kwargs.setdefault('capsize', 0)
    kwargs.setdefault('ms', 4)

    if remove_50000:
        time_offset = 50000

    for inst in self.instruments:
        s = self if self._child else getattr(self, inst)
        label = f'{inst:10s} ({s.N})' if N_in_label else inst

        y = getattr(s, quantity)
        ye = getattr(s, quantity + '_err')

        if np.isnan(y).all() or np.isnan(ye).all():
            lines, *_ = ax.errorbar([], [], [],
                                    label=label, picker=True, **kwargs)
            continue

        lines, *_ = ax.errorbar(s.mtime - time_offset, y[s.mask], ye[s.mask],
                                label=label, picker=True, **kwargs)

        if show_masked:
            ax.errorbar(s.time[~s.mask] - time_offset, y[~s.mask], ye[~s.mask],
                        **kwargs, color=lines.get_color())

    if show_masked:
        ax.errorbar(self.time[~self.mask] - time_offset,
                    getattr(self, quantity)[~self.mask],
                    getattr(self, quantity + '_err')[~self.mask],
                    label='masked', fmt='x', ms=10, color='k', zorder=-2)

    ax.legend()
    ax.minorticks_on()

    if quantity == 'fwhm':
        ax.set_ylabel(f'FWHM [{self.units}]')
    elif quantity == 'bispan':
        ax.set_ylabel(f'BIS [{self.units}]')
    elif quantity == 'rhk':
        ax.set_ylabel("$\log$ R'$_{HK}")

    if remove_50000:
        ax.set_xlabel('BJD - 2450000 [days]')
    else:
        ax.set_xlabel('BJD - 2400000 [days]')

    if config.return_self:
        return self
    else:
        return fig, ax


plot_fwhm = partialmethod(plot_quantity, quantity='fwhm')
plot_bis = partialmethod(plot_quantity, quantity='bispan')
plot_rhk = partialmethod(plot_quantity, quantity='rhk')


def gls(self, ax=None, label=None, fap=True, picker=True, **kwargs):
    if self.N == 0:
        if self.verbose:
            logger.error('no data to compute gls')
        return

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    gls = LombScargle(self.mtime, self.mvrad, self.msvrad)
    freq, power = gls.autopower(maximum_frequency=1.0, samples_per_peak=10)
    ax.semilogx(1/freq, power, picker=picker, label=label, **kwargs)

    if fap:
        ax.axhline(gls.false_alarm_level(0.01),
                   color='k', alpha=0.2, zorder=-1)

    ax.set(xlabel='Period [days]', ylabel='Normalized power', ylim=(0, None))
    ax.minorticks_on()

    if label is not None:
        ax.legend()


    if config.return_self:
        return self
    else:
        return fig, ax


def gls_quantity(self, quantity, ax=None, fap=True, picker=True):
    if not hasattr(self, quantity):
        logger.error(f"cannot find '{quantity}' attribute")
        return

    if ax is None:
        fig, ax = plt.subplots(1, 1, constrained_layout=True)
    else:
        fig = ax.figure

    t = self.mtime
    y = getattr(self, quantity)[self.mask]
    ye = getattr(self, quantity + '_err')[self.mask]

    if np.isnan(y).any():
        if self.verbose:
            logger.warning(f'{quantity} contains NaNs, ignoring them')
        m = np.isnan(y)
        t = t[~m]
        y = y[~m]
        ye = ye[~m]

    gls = LombScargle(t, y, ye)
    freq, power = gls.autopower(maximum_frequency=1.0)
    ax.semilogx(1/freq, power, picker=picker)

    if fap:
        ax.axhline(gls.false_alarm_level(0.01),
                   color='k', alpha=0.2, zorder=-1)

    ax.set(xlabel='Period [days]', ylabel='Normalized power', ylim=(0, None))
    ax.minorticks_on()

    if config.return_self:
        return self
    else:
        return fig, ax


gls_fwhm = partialmethod(gls_quantity, quantity='fwhm')
gls_bis = partialmethod(gls_quantity, quantity='bispan')
gls_rhk = partialmethod(gls_quantity, quantity='rhk')



import os, sys
import numpy as np
import matplotlib.pyplot as plt

from .setup_logger import logger

# ESPRESSO ADC issues
from .utils import ESPRESSO_ADC_issues

def ADC_issues(self, plot=True, check_keywords=True):
    instruments = self._check_instrument('ESPRESSO')
    
    if len(instruments) < 1:
        if self.verbose:
            logger.error(f"no data from ESPRESSO")
            logger.info(f'available: {self.instruments}')
    

    affected_file_roots = ESPRESSO_ADC_issues()
    file_roots = [os.path.basename(f).replace('.fits', '') for f in self.raw_file]
    intersect = np.in1d(file_roots, affected_file_roots)

    total_affected = intersect.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "affected by ADC issues")

    self.mask[intersect] = False
    self._propagate_mask_changes()

    if plot:
        self.plot(show_masked=True)

    return intersect

# ESPRESSO cryostat issues
from .utils import ESPRESSO_cryostat_issues

def blue_cryostat_issues(self, plot=True):
    instruments = self._check_instrument('ESPRESSO')
    
    if len(instruments) < 1:
        if self.verbose:
            logger.error(f"no data from ESPRESSO")
            logger.info(f'available: {self.instruments}')

    affected_file_roots = ESPRESSO_cryostat_issues()
    file_roots = [os.path.basename(f).replace('.fits', '') for f in self.raw_file]
    intersect = np.in1d(file_roots, affected_file_roots)

    total_affected = intersect.sum()

    if self.verbose:
        n = total_affected
        logger.info(f"there {'are'[:n^1]}{'is'[n^1:]} {n} frame{'s'[:n^1]} "
                     "affected by blue cryostat issues")

    self.mask[intersect] = False
    self._propagate_mask_changes()

    if plot:
        self.plot(show_masked=True)

    return intersect


def known_issues(self, plot=False, **kwargs):
    """ Identify and mask known instrumental issues (ADC and blue cryostat for ESPRESSO)

    Args:
        plot (bool, optional): Whether to plot the masked points.
    """
    adc = ADC_issues(self, plot, **kwargs)
    cryostat = blue_cryostat_issues(self, plot)
    return adc | cryostat



import os
import tarfile
import collections
import numpy as np
from dace_query import DaceClass
from dace_query.spectroscopy import SpectroscopyClass, Spectroscopy as default_Spectroscopy
from .setup_logger import logger
from .utils import create_directory, all_logging_disabled, stdout_disabled


def load_spectroscopy():
    if 'DACERC' in os.environ:
        dace = DaceClass(dace_rc_config_path=os.environ['DACERC'])
        return SpectroscopyClass(dace_instance=dace)
    # elif os.path.exists(os.path.expanduser('~/.dacerc')):
    return default_Spectroscopy

def get_arrays(result, latest_pipeline=True, ESPRESSO_mode='HR11', verbose=True):
    arrays = []
    instruments = list(result.keys())

    # if verbose:
    #     if latest_pipeline:
    #         logger.info('selecting latest pipeline version')

    for inst in instruments:
        pipelines = list(result[inst].keys())

        if 'ESPRESSO' in inst:
            if any(ESPRESSO_mode in pipe for pipe in pipelines):
                # if verbose:
                #     logger.info(f'selecting mode {ESPRESSO_mode} for ESPRESSO')
                i = [i for i, pipe in enumerate(pipelines) if ESPRESSO_mode in pipe][0]
                pipelines = [pipelines[i]]
            else:
                if verbose:
                    logger.warning(f'no observations for requested ESPRESSO mode ({ESPRESSO_mode})')

        if latest_pipeline:
            pipelines = [pipelines[0]]

        for pipe in pipelines:
            modes = list(result[inst][pipe].keys())
            for mode in modes:
                if 'rjd' not in result[inst][pipe][mode]:
                    logger.error(f"No 'rjd' key for {inst} - {pipe}")
                    raise ValueError

                arrays.append(
                    ((inst, pipe, mode), result[inst][pipe][mode])
                )

    return arrays

def get_observations(star, instrument=None, save_rdb=False, verbose=True):
    Spectroscopy = load_spectroscopy()
    try:
        with stdout_disabled(), all_logging_disabled():
            result = Spectroscopy.get_timeseries(target=star,
                                                sorted_by_instrument=True,
                                                output_format='numpy')
    except TypeError:
        if instrument is None:
            msg = f'no observations for {star}'
        else:
            msg = f'no {instrument} observations for {star}'
        raise ValueError(msg) from None

    # defaultdict --> dict
    if isinstance(result, collections.defaultdict):
        result = dict(result)
    for inst in result.keys():
        for pipe in result[inst].keys():
            for mode in result[inst][pipe].keys():
                if isinstance(result[inst][pipe][mode], collections.defaultdict):
                    result[inst][pipe][mode] = dict(result[inst][pipe][mode])
            if isinstance(result[inst][pipe], collections.defaultdict):
                    result[inst][pipe] = dict(result[inst][pipe])
        if isinstance(result[inst], collections.defaultdict):
            result[inst] = dict(result[inst])
    #

    instruments = list(result.keys())
    if instrument is not None:
        # select only the provided instrument (if it's there)
        instruments = [inst for inst in instruments if instrument in inst]

    if len(instruments) == 0:
        if instrument is None:
            msg = f'no observations for {star}'
        else:
            msg = f'no {instrument} observations for {star}'
        raise ValueError(msg)

    # sort pipelines, being extra careful with HARPS pipeline names
    # (i.e. ensure that 3.0.0 > 3.5)
    def cmp(a, b):
        if a[0] in ('3.5', '3.5 EGGS') and b[0] == '3.0.0':
            return -1
        if b[0] in ('3.5', '3.5 EGGS') and a[0] == '3.0.0':
            return 1

        if a[0] == b[0]:
            return 0
        elif a[0] > b[0]:
            return 1
        else:
            return -1

    from functools import cmp_to_key
    new_result = {}
    for inst in instruments:
        new_result[inst] = dict(sorted(result[inst].items(),
                                       key=cmp_to_key(cmp), reverse=True))

    if verbose:
        logger.info('RVs available from')
        with logger.contextualize(indent='  '):
            _inst = ''
            for inst in instruments:
                pipelines = list(new_result[inst].keys())
                for pipe in pipelines:
                    mode = list(new_result[inst][pipe].keys())[0]
                    N = len(new_result[inst][pipe][mode]['rjd'])
                    # LOG
                    if inst == _inst:
                        logger.info(f'{" ":>12s} └ {pipe:10s} ({N} observations)')
                    else:
                        logger.info(f'{inst:>12s} ├ {pipe:10s} ({N} observations)')
                    _inst = inst

    return new_result


def check_existing(output_directory, files, type):
    existing = [
        f.partition('_')[0] for f in os.listdir(output_directory)
        if type in f
    ]
    missing = []
    for file in files:
        if any(other in file for other in existing):
            continue
        missing.append(file)
    return np.array(missing)

def download(files, type, output_directory):
    
    Spectroscopy = load_spectroscopy()
    # with stdout_disabled(), all_logging_disabled():
    Spectroscopy.download_files(files, file_type=type,
                                output_directory=output_directory)

def extract_fits(output_directory):
    file = os.path.join(output_directory, 'spectroscopy_download.tar.gz')
    tar = tarfile.open(file, "r")
    files = []
    for member in tar.getmembers():
        if member.isreg():  # skip if the TarInfo is not a file
            member.name = os.path.basename(member.name)  # remove the path
            tar.extract(member, output_directory)
            files.append(member.name)
    os.remove(file)
    return files


def do_download_ccf(raw_files, output_directory, clobber=False, verbose=True):
    raw_files = np.atleast_1d(raw_files)

    create_directory(output_directory)

    # check existing files to avoid re-downloading
    if not clobber:
        raw_files = check_existing(output_directory, raw_files, 'CCF')

    # any file left to download?
    if raw_files.size == 0:
        if verbose:
            logger.info('no files to download')
        return

    if verbose:
        n = raw_files.size
        logger.info(f"Downloading {n} CCFs into '{output_directory}'...")

    download(raw_files, 'ccf', output_directory)

    if verbose:
        logger.info('Extracting .fits files')

    extract_fits(output_directory)


def do_download_s1d(raw_files, output_directory, clobber=False, verbose=True):
    raw_files = np.atleast_1d(raw_files)

    create_directory(output_directory)

    # check existing files to avoid re-downloading
    if not clobber:
        raw_files = check_existing(output_directory, raw_files, 'S1D')

    # any file left to download?
    if raw_files.size == 0:
        if verbose:
            logger.info('no files to download')
        return

    if verbose:
        n = raw_files.size
        logger.info(f"Downloading {n} S1Ds into '{output_directory}'...")

    download(raw_files, 's1d', output_directory)

    if verbose:
        logger.info('Extracting .fits files')

    extract_fits(output_directory)


def do_download_s2d(raw_files, output_directory, clobber=False, verbose=True):
    raw_files = np.atleast_1d(raw_files)

    create_directory(output_directory)

    # check existing files to avoid re-downloading
    if not clobber:
        raw_files = check_existing(output_directory, raw_files, 'S2D')

    # any file left to download?
    if raw_files.size == 0:
        if verbose:
            logger.info('no files to download')
        return

    if verbose:
        n = raw_files.size
        logger.info(f"Downloading {n} S2Ds into '{output_directory}'...")

    download(raw_files, 's2d', output_directory)

    if verbose:
        logger.info('Extracting .fits files')

    extracted_files = extract_fits(output_directory)
    return extracted_files



import numpy as np


def wmean(a, e):
    """Weighted mean of array `a`, with uncertainties given by `e`.
    The weighted mean is calculated using 1/e**2 as weights.

    Args:
        a (array): Array containing data
        e (array): Uncertainties on `a`
    """
    return np.average(a, weights=1 / e**2)


def rms(a):
    """ Root mean square of array `a`

    Args:
        a (array): Array containing data
    """
    return np.sqrt((a**2).mean())


def wrms(a, e):
    """ Weighted root mean square of array `a`, with uncertanty given by `e`.
    The weighted rms is calculated using the weighted mean, where the weights
    are equal to 1/e**2.
    
    Args:
        a (array): Array containing data 
        e (array): Uncertainties on `a`
    """
    w = 1 / e**2
    return np.sqrt(np.sum(w * (a - np.average(a, weights=w))**2) / sum(w))



import pytest
import os

def test_import():
    import arvi
    from arvi import RV


def test_import_object():
    from arvi import _51Peg

    if os.path.exists(os.path.expanduser('~/.dacerc')):
        from arvi import HD10180
    else:
        with pytest.raises(ValueError):
            from arvi import HD10180




__version__ = '0.1.2'
__all__ = []

from .timeseries import RV

_ran_once = False

def __getattr__(name: str):
    global _ran_once  # can't do it any other way :(
    if _ran_once:
        return RV(name)
    else:
        _ran_once = True



from functools import partial
import numpy as np
from astropy.timeseries import LombScargle
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from matplotlib.backends.backend_pdf import PdfPages

from .setup_logger import logger


sine_line = None
residual_gls = None
def sine_picker(event, self, fig, ax, ax1):
    from .timeseries import fit_sine
    global sine_line, residual_gls
    if sine_line is not None:
        sine_line[0].remove()
    if residual_gls is not None:
        residual_gls[0].remove()
    xdata, ydata = event.artist.get_data()
    ind = event.ind
    period = xdata[ind][0]
    p, sine = fit_sine(self.mtime, self.mvrad, self.msvrad, period=period)
    tt = np.linspace(self.mtime.min(), self.mtime.max(), 100)
    tt -= 50000
    sine_line = ax1.plot(tt, sine(tt), 'k')
    #
    f, p = LombScargle(self.mtime, self.mvrad - sine(self.mtime), self.msvrad).autopower()
    residual_gls = ax.semilogx(1/f, p, 'r')
    fig.canvas.draw_idle()


def report(self, save=None):
    # size = A4
    size = 8.27, 11.69
    fig = plt.figure(figsize=size, constrained_layout=True)
    gs = gridspec.GridSpec(5, 3, figure=fig, height_ratios=[2, 2, 1, 1, 0.1])

    # first row, all columns
    ax1 = plt.subplot(gs[0, :])

    title = f'{self.star}'
    ax1.set_title(title, loc='left', fontsize=14)
    # ax1.set_title(r"\href{http://www.google.com}{link}", color='blue',
    #               loc='center')

    if self._did_adjust_means:
        title = '(instrument means subtracted) '
    else:
        title = ''
    title += f'V={self.simbad.V}, {self.simbad.sp_type}'
    ax1.set_title(title, loc='right', fontsize=12)

    self.plot(ax=ax1, N_in_label=True, tooltips=False, remove_50000=True)


    ax1.legend().remove()
    legend_ax = plt.subplot(gs[1, -1])
    legend_ax.axis('off')
    leg = plt.legend(*ax1.get_legend_handles_labels(),
                     prop={'family': 'monospace'})
    legend_ax.add_artist(leg)
    second_legend = f'rms  : {self.rms:.2f} {self.units}\n'
    second_legend += f'error: {self.error:.2f} {self.units}'
    legend_ax.legend([],
                     title=second_legend,
                     loc='lower right', frameon=False,
                     prop={'family': 'monospace'})

    ax2 = plt.subplot(gs[1, :-1])
    self.gls(ax=ax2, picker=True)

    ax3 = plt.subplot(gs[2, :-1])
    self.plot_fwhm(ax=ax3, tooltips=False, remove_50000=True)
    ax3.legend().remove()
    ax3p = plt.subplot(gs[2, -1])
    self.gls_fwhm(ax=ax3p, picker=False)

    ax4 = plt.subplot(gs[3, :-1])
    self.plot_bis(ax=ax4, tooltips=False, remove_50000=True)
    ax4.legend().remove()
    ax4p = plt.subplot(gs[3, -1])
    self.gls_bis(ax=ax4p, picker=False)


    if save is None:
        fig.canvas.mpl_connect(
            'pick_event',
            partial(sine_picker, self=self, fig=fig, ax=ax2, ax1=ax1))

    if save is not None:
        if save is True:
            save = f'report_{"".join(self.star.split())}.pdf'

        with PdfPages(save) as pdf:
            pdf.attach_note('hello', positionRect=[5, 15, 20, 30])

            if self.verbose:
                logger.info(f'saving to {save}')
            pdf.savefig(fig)
        plt.close('all')
        # os.system(f'evince {save} &')